#include "OitoDamas.h"
#include <stdio.h>

// dados de instncia
int CTorres::N=4;



CTorres::CTorres(void)
{
}

CTorres::~CTorres(void)
{
}

// Cria um objecto que e uma copia deste
TProcuraConstrutiva* CTorres::Duplicar(void)
{
	CTorres* clone = new CTorres;
	clone->torres = torres;
	clone->inutilizadas = inutilizadas;
	clone->ultLinha = ultLinha;
	clone->ultColuna = ultColuna;
	clone->tabuleiro = tabuleiro;
	clone->custo = custo;
	return clone;
}

// Coloca o objecto no estado inicial da procura
void CTorres::SolucaoVazia(void) {
	int resetValue = 0;
	torres = inutilizadas = 0;
	ultLinha = ultColuna = 0;
	tabuleiro.Count(N * N); 
	_SETALTTORRE(resetValue, 7); // todas as torres podem ser colocadas inicialmente
	tabuleiro.Reset(resetValue);
}


// Coloca em sucessores a lista de objectos sucessores (sao alocados neste metodo e tem de ser apagados)
// O custo nao necessita de ser preenchido, caso seja sempre unitario
void CTorres::Sucessores(TVector<TProcuraConstrutiva*>&sucessores, TVector<int>&custo)
{
	// adicionar sucessores apenas aps o ltimo sucessor 
	CTorres* sucessor = (CTorres*)Duplicar();
	for(int i=ultLinha;i<N;i++)
		for(int j=(i==ultLinha?ultColuna:0);j<N;j++)
			for (int l = 1; l <= 3; l++) 
				if (sucessor->ColocarTorre(i, j, l)) {
#ifdef HASHSIZE 
					// hashtable
					CTorres* aux = (CTorres*)sucessor->Duplicar();
					aux->Normalizar();
					if (hashTableID(aux->hash(), aux->tabuleiro) >= 0) {
						sucessores.Add(sucessor);
						custo.Add(sucessor->inutilizadas - inutilizadas); // desperdcio de casas neste movimento
}
					else // no colocar o sucessor, j existe
						delete sucessor;
#else
					sucessores.Add(sucessor);
					custo.Add(sucessor->inutilizadas - inutilizadas); // desperdcio de casas neste movimento
#endif
					sucessor = (CTorres*)Duplicar();
				}
	delete sucessor;
	TProcuraConstrutiva::Sucessores(sucessores,custo);
}

// normalizar este tabuleiro, antes de utilizar a hashtable
void CTorres::Normalizar()
{
	int peso = Peso();
	TVector<int> ordem, invOrdem;
//	Debug();
//	printf("\nPeso: %d", peso);
	// eixo horizontal/vertical/diagonal
	for (int i = 0; i < 3; i++) {
		Inverter(0);
		if (peso < Peso()) {
			// reverter
			Inverter(0);
		}
		else {
//			Debug();
//			printf("\nPeso: %d", peso);
		}
	}
	// ordem das cores das torres
	int aux;
	for (int i = 0; i < tabuleiro.Count() && ordem.Count()<3; i++) {
		if ((aux=_TORRE(tabuleiro[i])) > 0) {
			if (ordem.Find(aux) < 0)
				ordem.Add(aux);
		}
	}
	if (ordem.Count() > 0) {
		invOrdem.Count(3);
		invOrdem.Reset(1);
		for (int i = 0; i < ordem.Count(); i++)
			invOrdem[ordem[i] - 1] = i + 1;
	}
	for (int i = 0; i < tabuleiro.Count() && ordem.Count() < 3; i++) {
		if ((aux = _TORRE(tabuleiro[i])) > 0) {
			aux = invOrdem[aux - 1];
			_SETTORRE(tabuleiro[i], aux);
		}
		_SETALTTORRE(tabuleiro[i], 0);
	}
//	Debug();
}

// peso de um tabuleiro, para colocar na verso normalizada de menor peso
int CTorres::Peso() 
{
	int peso = 0;
	for (int i = 0; i < N; i++)
		for (int j = 0; j < N; j++)
			peso += (_TORRE(Casa(i, j)) > 0) * (i * N + j + 1);
	return peso;
}

// inverte o tabuleiro no eixo especificado
void CTorres::Inverter(int eixo)
{
	if(eixo==0 || eixo==1)
		for (int i = 0; i < N / 2; i++) {
			for (int j = 0; j < N; j++) 
				if(eixo==0) { // horizontal
					int aux = Casa(i, j);
					Casa(i, j) = Casa(N - 1 - i, j);
					Casa(N - 1 - i, j) = aux;
				}
				else if (eixo == 1) { // vertical
					int aux = Casa(j, i);
					Casa(j, i) = Casa(j, N - 1 - i);
					Casa(j, N - 1 - i) = aux;

				}
		}
	else { // diagonal
		for (int i = 0; i < N ; i++) 
			for (int j = i + 1; j < N; j++) {
				int aux = Casa(i, j);
				Casa(i, j) = Casa(j, i);
				Casa(j, i) = aux;
			}
	}
}

bool CTorres::ColocarTorre(int linha, int coluna, int cor)
{
	int alt, resetCor, coresLinha, coresColuna;
	if (cor == 0) { // apenas aplicar restries das peas colocadas devido  colocao de nova pea
		coresLinha = coresColuna = 0;
		for (int i = 0; i < N; i++) {
			alt = _TORRE(Casa(i, coluna));
			if (alt)
				coresLinha |= 1 << (alt - 1);
			alt = _TORRE(Casa(linha, i));
			if (alt)
				coresColuna |= 1 << (alt - 1);
		}
		cor = _TORRE(Casa(linha, coluna));
		// ver qual o reset a fazer nas cores da linha/coluna
		resetCor = coresLinha | coresColuna | (1 << (cor - 1));
		// remover a possibilidade desta torre (e outras vizinhas) estar na mesma linha/coluna
		for (int i = 0; i < N; i++) {
			alt = _ALTTORRE(Casa(i, coluna));
			if (alt & resetCor) {
				alt &= ~resetCor;
				_SETALTTORRE(Casa(i, coluna), alt);
			}
			alt = _ALTTORRE(Casa(linha, i));
			if (alt & resetCor) {
				alt &= ~resetCor;
				_SETALTTORRE(Casa(linha, i), alt);
			}
		}
	} else if (_TORRE(Casa(linha, coluna)) == 0 && _ALTTORRE(Casa(linha, coluna))&(1<<(cor-1))) {
		// verificar se a colocao de uma torre nesta casa, cria uma violao de vizinhos (coluna com linha com torre com mesma cor)
		coresLinha = coresColuna = 0;
		for (int i = 0; i < N; i++) {
			alt = _TORRE(Casa(i, coluna));
			if (alt)
				coresLinha |= 1 << (alt - 1);
			alt = _TORRE(Casa(linha, i));
			if (alt)
				coresColuna |= 1 << (alt - 1);
		}
		if (!(coresLinha & coresColuna)) { // se intersectar, no d
			torres++;
			_SETTORRE(Casa(linha, coluna), cor);
			_SETALTTORRE(Casa(linha, coluna), 0);
			// ver qual o reset a fazer nas cores da linha/coluna
			resetCor = coresLinha | coresColuna | (1 << (cor - 1));
			// remover a possibilidade desta torre (e outras vizinhas) estar na mesma linha/coluna
			for (int i = 0; i < N; i++) {
				alt = _ALTTORRE(Casa(i, coluna));
				if (alt & resetCor) {
					alt &= ~resetCor;
					_SETALTTORRE(Casa(i, coluna), alt);
				}
				alt = _ALTTORRE(Casa(linha, i));
				if (alt & resetCor) {
					alt &= ~resetCor;
					_SETALTTORRE(Casa(linha, i), alt);
				}
			}
			// reaplicar as restries em peas j colocadas
			for (int i = 0; i < N; i++) {
				alt = _TORRE(Casa(i, coluna));
				if (i!=linha && alt) 
					ColocarTorre(i, coluna, 0);
				alt = _TORRE(Casa(linha, i));
				if (i!=coluna && alt) 
					ColocarTorre(linha, i, 0);
			}
			// calcular as casas inutilizadas
			inutilizadas = -torres; // no contar com as torres
			for (int i = 0; i < N; i++)
				for (int j = 0; j < N; j++)
					if (!_ALTTORRE(Casa(i, j)))
						inutilizadas++;
			ultLinha = linha;
			ultColuna = coluna;
			return true;
		}
	}
	return false;
}

// Escrever informacao de debug sobre o objecto currente
// (utilizar variavel TProcuraConstrutiva::debug para seleccionar o detalhe pretendido)
void CTorres::Debug(void)
{
	for(int i=0;i<N;i++) {
		NovaLinha();
		for(int j=0;j<N;j++) 
			printf("%c ", _TORRE(Casa(i, j))?'A'+ _TORRE(Casa(i, j))-1:'+');
	}
	if (N <= 4) {
		// mostrar mais detalhe
		for (int i = 0; i < N; i++) {
			NovaLinha();
			for (int j = 0; j < N; j++)
				printf("%c%c%c ",
					_ALTTORRE(Casa(i, j)) & 1 ? 'a' : ' ',
					_ALTTORRE(Casa(i, j)) & 2 ? 'b' : (_TORRE(Casa(i, j)) > 0 ? 'A' + _TORRE(Casa(i, j)) - 1 : '+'),
					_ALTTORRE(Casa(i, j)) & 4 ? 'c' : ' ');
		}
		NovaLinha();
		printf("%d torres %d inutilizadas (hash %u, avaliacao %d, heuristica %d)",
			torres, inutilizadas, hash(), Avaliar(), Heuristica());
	}
	else {
		NovaLinha();
		printf("%d torres %d inutilizadas",
			torres, inutilizadas);
	}

	NovaLinha();
}

unsigned int CTorres::hash()
{
	unsigned int hash = 3513857;;
	// utilizar o tabuleiro
	for (int i = 0; i < tabuleiro.Count(); i++) {
		hash ^= tabuleiro[i] ^ 5351;
		hash ^= (tabuleiro[i] << 5) ^ 8221541;
		hash = (hash << 9) | (hash >> 2);
		hash += 463131;
	}
	return hash;
}


// calcular das casas que ainda no tm torres nem esto inutilizadas, quais dessas no podem ser utilizadas
int CTorres::Heuristica(void) { 
	int valorLinhas = 0, valorColunas=0;
	avaliacoes++; 
	for (int i = 0; i < N; i++) {
		int alt = 0, validas=0, ncores=0;
		// verificar se nesta linha, quantas possibilidades de torres existe
		for (int j = 0; j < N; j++) {
			alt |= _ALTTORRE(Casa(i, j));
			if (_ALTTORRE(Casa(i, j)))
				validas++;
		}
		while (alt) {
			if (alt & 1)
				ncores++;
			alt >>= 1;
		}
		if (ncores < validas)
			valorLinhas += validas - ncores; // estas casas no iro ser utilizadas de certeza
		// repetir para colunas
		alt = validas = ncores = 0;
		for (int j = 0; j < N; j++) {
			alt |= _ALTTORRE(Casa(j, i));
			if (_ALTTORRE(Casa(j, i)))
				validas++;
		}
		while (alt) {
			if (alt & 1)
				ncores++;
			alt >>= 1;
		}
		if (ncores < validas)
			valorColunas += validas - ncores; // estas casas no iro ser utilizadas de certeza
	}

	// retornar o maior valor que se detecta, em linhas ou colunas
	return valorColunas > valorLinhas ? valorColunas : valorLinhas;
}

// Retorna verdade caso o estado actual seja um estado objectivo
// neste caso, no  possvel colocar mais torres
bool CTorres::SolucaoCompleta(void)
{
	return inutilizadas + torres == N * N;
}


// e gerada uma nova solucao aleatoriamente.
void CTorres::NovaSolucao(void)
{
	SolucaoVazia();

	Completar();
}

// preencher o resto das casas pendentes de forma aleatria
void CTorres::Completar()
{
	static TVector<int> linha, coluna, cor, id;

	// montar vetores apenas uma s vez por cada instncia
	if (linha.Count() != N * N * 3) {
		linha.Count(0);
		coluna.Count(0);
		cor.Count(0);
		id.Count(0);
		for (int i = 0; i < N; i++)
			for (int j = 0; j < N; j++)
				for (int l = 1; l <= 3; l++) {
					linha.Add(i);
					coluna.Add(j);
					cor.Add(l);
					id.Add(id.Count());
				}
	}

	id.RandomOrder();

	for (int i = 0; i < id.Count() && N * N>torres + inutilizadas; i++)
		ColocarTorre(linha[id[i]], coluna[id[i]], cor[id[i]]);
}


// Retorna o valor da solucao completa actual.
int CTorres::Avaliar(void)
{
	TProcuraMelhorativa::Avaliar();
	// inutilizadas na soluo completa, mas se existirem casas que possam ser preenchidas, contam como N*N
	custo = inutilizadas + N*N * (N * N - torres - inutilizadas);
	return custo;
}

// Operador de vizinhanca (apenas em solucoes completas)
void CTorres::Vizinhanca(TVector<TProcuraMelhorativa*>&vizinhos)
{
	// chamar simplesmente o operador de mutao N vezes

	for (int i = 0; i < N; i++) {
		CTorres* vizinho = (CTorres*)Duplicar();
		vizinho->Mutar();
		vizinhos.Add(vizinho);
	}
	TProcuraMelhorativa::Vizinhanca(vizinhos);
}

// muta o estado actual (de acordo com a parametrizacao global)
void CTorres::Mutar(void)
{
	// remover algumas peas colocadas aleatoriamente, e completar de seguida
	TVector<int> linha, coluna, cor;
	for(int i=0;i<N;i++)
		for(int j=0;j<N;j++)
			if (_TORRE(Casa(i, j)) > 0) {
				linha.Add(i);
				coluna.Add(j);
				cor.Add(_TORRE(Casa(i, j)));
			}
	do {
		int id = TRand::rand() % linha.Count();
		linha[id] = linha.Last();
		coluna[id] = coluna.Last();
		cor[id] = cor.Last();
		linha.Pop();
		coluna.Pop();
		cor.Pop();
	} while (TRand::rand() % 2 && linha.Count() > 1); // pode parar aqui com 50% de probabilidade

	SolucaoVazia();
	for (int i = 0; i < linha.Count(); i++)
		ColocarTorre(linha[i], coluna[i], cor[i]);

	// recolocar o resto da soluo
	Completar();

	TProcuraMelhorativa::Mutar();
}

// cruza os dois elementos a e b, colocando o resultado neste
void CTorres::Cruzamento(TProcuraMelhorativa *a, TProcuraMelhorativa *b)
{
	CTorres& pai = *((CTorres*)a);
	CTorres& mae = *((CTorres*)b);
	SolucaoVazia();
	// Executar os passos por ordem
	// 1. coloca as torres que no pai e me esto na mesma casa, com a cor do pai
	for (int i = 0; i < N; i++)
		for (int j = 0; j < N; j++)
			if (_TORRE(pai.Casa(i, j)) > 0 && _TORRE(mae.Casa(i, j)) > 0)
				ColocarTorre(i, j, _TORRE(pai.Casa(i, j)));

	// 2. procurar colocar as torres nas casas da me e cores da me
	for (int i = 0; i < N; i++)
		for (int j = 0; j < N; j++)
			if (_TORRE(mae.Casa(i, j)) > 0)
				ColocarTorre(i, j, _TORRE(mae.Casa(i, j)));

	// 3. fazer o mesmo com o pai
	for (int i = 0; i < N; i++)
		for (int j = 0; j < N; j++)
			if (_TORRE(pai.Casa(i, j)) > 0)
				ColocarTorre(i, j, _TORRE(pai.Casa(i, j)));

	// recolocar o resto da soluo
	Completar();

	TProcuraMelhorativa::Cruzamento(a,b);
}

// retorna a distancia deste elemento ao elemento a
// (necessario para manter os elementos diversos)
int CTorres::Distancia(TProcuraMelhorativa*a)
{
	int distancia = 0;
	CTorres& estado = *((CTorres*)a);
	// contar casas que tm torres ou no tm torres (se diferente  que incrementa a distncia)
	for (int i = 0; i < N; i++)
		for (int j = 0; j < N; j++)
			if(_TORRE(Casa(i,j))>0 && _TORRE(estado.Casa(i,j))==0 || 
				_TORRE(Casa(i, j)) == 0 && _TORRE(estado.Casa(i, j)) > 0) 
			distancia++;
	return distancia;
}

#ifdef HASHSIZE

TVector<int> CTorres::hashTable[HASHSIZE];

void CTorres::ResetHashTable()
{
	for (int i = 0; i < HASHSIZE; i++)
		hashTable[i].Count(0);
}

int CTorres::hashTableID(unsigned int hash, TVector<int>&tabuleiro)
{
	unsigned int id = hash % HASHSIZE;

	if (hashTable[id].Equal(tabuleiro))
		return -1;
	else {
		hashTable[id] = tabuleiro;
		return id;
	}
}

#endif
